/* $Workfile: T2mRaw30.c $       $Revision: 19 $      $Modtime: 9/12/03 8:54a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2MRAW30.C

    A simple example console application demonstrating the use of the 
    Raw Data Sender and the Raw Data Receiver with the OTX Driver.

    The applications works with the THOR-2-PCI / THOR-8-PCI board equipped
    with an ASM daughter board, such as:

      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM,
      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM-CST,
      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM-PRO,
      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM-EX,
      Thor-2-PCI / Thor-8-PCI + Vidar-55x4-ASM,

    or Thor-2-PCMCIA series boards,

      Thor-2-PCMCIA, 
      Thor-2-PCMCIA-Plus, 
      Thor-2-PCMCIA-Ex, 
      Thor-2-PCMCIA-Pro,
      Thor-2-PCMCIA-CST

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 E1T1 devices in E1 mode. Li0 should be looped to Li1 with an external
    loop cable. The LI Internal HDLC controller is used to send HDLC frames from
    one Line Interface to be received by the other.

    The DSP0 and DSP1 are loaded with the OTX_DSP_SPM_OTXHDLC program.

    30 64kpbs timeslots on Li0 are connected to DSP0. 30 RawDataSender devices are configured
    DSP0. The RawDataSender devices are continously sending data from a static array of data 
    defined in this application.

    30 64kpbs timeslots on Li1 are connected to DSP1. 30 RawDataReceiver devices are configured
    DSP1.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(_WIN32_WCE)
#include "Lib/WindowsCe/OsaWindowsCe.h"
#else   // #if defined(_WIN32_WCE)
#include <time.h>
#include <assert.h>
#endif  // #if defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"

#define DEMO_DEFAULT_LOG_FILE "T2mRaw30.log"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_RAWDATA_DEVICES_MAX (2*32)

#define DEMO_RAW_TX_BUF_SIZE 1024
//#define DEMO_RAW_TX_BUF_SIZE 2000

//#define DEMO_LINK_WITH_LIBCMT 1

#if defined(__linux__)
#define MAX_PATH 260    // linux doesn't define MAX_PATH
#endif

#ifdef DEMO_LINK_WITH_LIBCMT
#include <process.h>
#endif

typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[MAX_PATH];
  char            logFileName[MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  int             boardNo;
  int             coDsp;
  int             coRawDevs;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
OTX_INT32   g_cbLiHdlcSenderFrame  = 10;
OTX_INT32   g_cbDspHdlcSenderFrame = 5;
OTX_INT32   g_coHdlcSenderFrames   = 1;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_INT32   g_iRawRangeEnd;
OTX_BOOL    g_bStatsThreadActive = OTX_FALSE;


//CRITICAL_SECTION g_analyzeCritSec;
//CRITICAL_SECTION g_demoLogfCritSec;


struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hVidarAsmController0;
  OTX_HANDLE hLi[2];
  OTX_HANDLE hDsp[4];
} g_hPhysDevs;



struct {
  // External Pipes (The T1/E1 span) 
  OTX_HANDLE hLiIn;
  OTX_HANDLE hLiOut;

} g_hPipes[2];


struct sRawDataSender {
  OTX_HANDLE hPipe;

  OTX_HANDLE hHandle;
  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_UINT32 nSourceId;

  //DWORD nThreadId;
  //HANDLE hThread;
  OTX_INT32 nDsp;
  OTX_INT32 nLi;

  OTX_UINT8   sqTxBuf[DEMO_RAW_TX_BUF_SIZE];
  //FILE       *pTxFile;
  OTX_INT32   cbTxBuf;
  OTX_INT32   cbSent;
  OTX_INT32   coUnderRunErrors;
};

struct sRawDataReceiver {
  OTX_HANDLE hPipe;

  OTX_HANDLE hHandle;
  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_UINT32 nSourceId;

  //DWORD nThreadId;
  //HANDLE hThread;
  OTX_INT32 nDsp;
  OTX_INT32 nLi;

  //FILE       *pRxFile;
  OTX_UINT8   sqRxBuf[2048];
  OTX_INT32   cbReceived;
};

struct {
  OTX_HANDLE hHdlcSender[2];
  OTX_HANDLE hHdlcReceiver[2];
  struct sRawDataSender mRawDataSender[DEMO_RAWDATA_DEVICES_MAX];
  struct sRawDataReceiver mRawDataReceiver[DEMO_RAWDATA_DEVICES_MAX];
} g_hLogicDevs;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;




/*******************************************************************************
*******************************************************************************/

/*******************************************************************************

  @func


*******************************************************************************/
void DemoInitalizeGlobalData(void)
{
  OTX_INT32  i;

  memset(g_hLogicDevs.mRawDataSender,    0, DEMO_RAWDATA_DEVICES_MAX*sizeof(struct sRawDataSender));
  memset(g_hLogicDevs.mRawDataReceiver,    0, DEMO_RAWDATA_DEVICES_MAX*sizeof(struct sRawDataReceiver));

  for (i = 0; i < DEMO_RAWDATA_DEVICES_MAX; i++) {
    // Device[ 0-31] DSP0, Li0; 
    // Device[32-63] DSP2, Li1; 
    g_hLogicDevs.mRawDataSender[i].nDsp      = (i/32) * 2; // DSP0, DSP2
    g_hLogicDevs.mRawDataSender[i].nLi       = i/32;       // Li0, Li1

    // Device[ 0-31] DSP1, Li1; 
    // Device[32-63] DSP3, Li0; 
    g_hLogicDevs.mRawDataReceiver[i].nDsp    = ((i/32) * 2) + 1; // DSP1, DSP3
    g_hLogicDevs.mRawDataReceiver[i].nLi     = (DEMO_RAWDATA_DEVICES_MAX - 1 - i)/32;       // Li1, Li0

    
  }
}


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the OTX board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.


  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_THOR_PCI:
      DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
      break;
    case OTX_DEVICE_THOR_PCMCIA:
      DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      exit(1);
      break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Open the two Line interfaces
  for (i = 0; i < 2; i++) {
    // Get handles to T1/E1 line interface devices #0 and #1
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }

                    
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
    // Get handles to the physcial DSP devices #0 and #1
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI) {

    // Get a handle to to the ASM Daughter Board
    OTX_UINT32   eDevType;
    OTX_ITERATOR itrChildTypes;
    OtxDrvIteratorBegin(&itrChildTypes);
    while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
      OTX_BOOL bOpenAsm = OTX_FALSE;
      switch (eDevType) {
      case OTX_DEVICE_VIDAR_5x_ASM:
        DemoLogf("Opening Vidar-5x4-ASM board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_CST:
        DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_V90:
        DemoLogf("Opening Vidar-5x4-ASM-V90 board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_EX:
        DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
        DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_55x4_ASM:
        DemoLogf("Opening Vidar-55x4-ASM board\n");
        bOpenAsm = OTX_TRUE;

        // Verify that the system supports the Vidar-55x4-ASM board
        {
          char szCpldBuildNumbers[256];
          char szBrdBuildNumber[16];
          OTX_INT32 nBoardBuildNo;
          
          if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI) {
            
            //  Verify firmware revision
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            if (strstr(szCpldBuildNumbers, "OTXCDC: 0761") || strstr(szCpldBuildNumbers, "OTXCDC: 0762")) {
              DemoLogf("Error: The firmware on the Thor-2-PCI must be upgraded to firmware build 5\n");
              DemoLogf("       (OTXCDC 0764) or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
              DemoLogf("       web site (www.odints.com)\n");
              DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
              DemoExitQuestion();
            }
            
            // Verify board build number
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            nBoardBuildNo = strtol(strtok(szBrdBuildNumber, "."), NULL, 0);
            DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
            if (nBoardBuildNo < 11) {
              DemoLogf("Error: The Thor-2-PCI board must be upgraded to build 11\n");
              DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
              DemoExitQuestion();
            }
            
          } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
            //  Verify firmware revision
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            if (strstr(szCpldBuildNumbers, "OTXCDC: 0771")) {
              DemoLogf("Error: The firmware on the Thor-8-PCI must be upgraded to firmware build 3\n");
              DemoLogf("       (OTXCDC 0772) or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
              DemoLogf("       web site (www.odints.com)\n");
              DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
              DemoExitQuestion();
            }
            
            // Verify board build number
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            nBoardBuildNo = strtol(szBrdBuildNumber, NULL, 0);
            DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
            if (nBoardBuildNo < 5) {
              DemoLogf("Error: The Thor-8-PCI board must be upgraded to build 5\n");
              DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
              DemoExitQuestion();
            }
            
          }
          
        }
        break;
      }
      if (bOpenAsm) {
        // Get a handle to to the ASM Daughter Board
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
    }
    
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
    }
  } else {
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
  }
  
  // Load DSP Program
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXHDLC\n", i);
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], OTX_DSP_SPM_OTXHDLC); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  for (i = 0; i < 2; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    // Configure Transceivers
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

  }

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable hThorController0\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

  // Enable Physical Devices
  for (i = 0; i < 2; i++) {
    DemoLogf("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    DemoLogf("g_hPhysDevs.hDsp[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}


void DemoCreateLogicalDevices(
/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

    Logical Devices can be classified into Control Device and Data Devices:

    - Control Device:  A logical device performing a control function. E.g. Event Queue.

    - Data Filter Device: A device implemented in Hardware, Firmware of Software
    which converts, processes, or manipulates serial data. Examples of data devices include
    Time-Space Switch, HDLC Controller, Tone Detector, Tone Generator, etc.

    Data Filter devices can be categorized into 3 basic types:

     - Data Source: Originates a Pipe or adds data to an existing Pipe. E.g. DTMF Tone Generator.

     - Data Sink: Terminates a Pipe or extracts data from a Pipe. E.g. DTMF Detector.

     - Data Bridge: Cross-connects data from one Pipe to another or combines data from
     multiple Pipes into one Pipe. E.g. cross-connect in TSS.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  // LI#0 and LI#1: Create Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
  for (i = 0; i < 2; i++) {
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hLi[i], OTX_LDEVICE_HDLC_SENDER,   g_hEventQueue, i, &g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hLi[i], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }


  // Create 30 RawData Sender devices hosted by DSP0 (and DSP2)
  for (i = 0; i < g_demoEnv.coRawDevs; i++) {
    OTX_INT32 nDsp = g_hLogicDevs.mRawDataSender[i].nDsp;

    if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
    // Create a notification Event and Event Queue for the Raw Data Sender
    result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataSender[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_hLogicDevs.mRawDataSender[i].hNotificationEvent, &g_hLogicDevs.mRawDataSender[i].hEventQueue); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hEventQueue, result);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hEventQueue, result);

    // Create Raw Data Sender
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_DATA_RAW_SENDER, g_hLogicDevs.mRawDataSender[i].hEventQueue, i, &g_hLogicDevs.mRawDataSender[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
    g_hLogicDevs.mRawDataSender[i].nSourceId = i; // Store sourceId so that we can easily print it from the Sender Thread
  }

  // Create 30 RawData Receiver devices hosted by DSP1 (and DSP3)
  for (i = 0; i < g_demoEnv.coRawDevs; i++) {
    OTX_UINT32 nCurrentFifoSize;
    OTX_UINT32 nNewFifoSize;
    OTX_INT32 nDsp = g_hLogicDevs.mRawDataReceiver[i].nDsp;

    if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
    // Create a notification Event and Event Queue for the Raw Data Receiver
    result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataReceiver[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_hLogicDevs.mRawDataReceiver[i].hNotificationEvent, &g_hLogicDevs.mRawDataReceiver[i].hEventQueue); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hEventQueue, result);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hEventQueue, result);

    // Create Raw Data Receiver
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_DATA_RAW_RECEIVER, g_hLogicDevs.mRawDataReceiver[i].hEventQueue, i, &g_hLogicDevs.mRawDataReceiver[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
    g_hLogicDevs.mRawDataReceiver[i].nSourceId = i; // Store sourceId so that we can easily print it from the Receiver Thread

    // Modify the FIFO size of the Raw Data Receiver
    result = OtxDrvGetAttributeValueUint32(
      g_hLogicDevs.mRawDataReceiver[i].hHandle, 
      OTX_ATTR_RAWDATA_RECEIVER_FIFO_SIZE, 
      &nCurrentFifoSize); 
    DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
    
    nNewFifoSize = 16*1024;
    DemoLogf("RawDataReceiver[%d]: Setting nFifoSize = %d bytes  (current size = %d bytes)\n", i, nNewFifoSize, nCurrentFifoSize);
    result = OtxDrvSetAttributeValueUint32(
      g_hLogicDevs.mRawDataReceiver[i].hHandle, 
      OTX_ATTR_RAWDATA_RECEIVER_FIFO_SIZE, 
      nNewFifoSize); 
    DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);

  }


}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.


*******************************************************************************/
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 nCapacity = 64;


  // Create pipes for the HDLC Logicial Device in the Line Interface (one for each direction)
  for (i = 0; i < 2; i++) {
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    
	{
    OTX_INT32 ts;
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      ts = 24; 
    } else {
      // E1 mode
      ts = 16; 
    }
    
    sqTimeSlotMask[ts] = 0xFF;
	}
    
    // Create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[i].hLiIn); DemoCheckRc(g_hPipes[i].hLiIn, result);
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[i].hLiOut); DemoCheckRc(g_hPipes[i].hLiOut, result);
    
    // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    result = OtxDrvConnectPipe(g_hPipes[i].hLiIn,  OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hLiIn, result);
    result = OtxDrvConnectPipe(g_hPipes[i].hLiOut, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hLiOut, result);
  }


  // CONNECT LOGICAL DEVICES WITHIN THE LIs
  // --------------------------------------
  for (i = 0; i < 2; i++) {
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hHdlcSender[i], -1, g_hPipes[i].hLiOut, -1); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);
    result = OtxDrvConnectLogicalDevice(g_hPipes[i].hLiIn,           -1, g_hLogicDevs.hHdlcReceiver[i], -1); DemoCheckRc(g_hPipes[i].hLiIn, result);
  }




  // Create the Pipes for DSP0->Li0 (30 RawDataSender pipes) device  0-31
  // Create the Pipes for DSP2->Li1 (30 RawDataSender pipes) device 32-63
  for (i = 0; i < g_demoEnv.coRawDevs; i++) {
    OTX_INT32 nTs = i%32;
    OTX_INT32 nDsp = g_hLogicDevs.mRawDataSender[i].nDsp;
    OTX_INT32 nLi = g_hLogicDevs.mRawDataSender[i].nLi;
        
    if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nTs] = 0xFF; 
    
    result = OtxDrvCreatePipe(nCapacity, &g_hLogicDevs.mRawDataSender[i].hPipe); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hPipe, result);
    
    // Connect pipe: Dsp -> Li 
    result = OtxDrvConnectPipe(g_hLogicDevs.mRawDataSender[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hPipe, result);
    result = OtxDrvConnectPipe(g_hLogicDevs.mRawDataSender[i].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hPipe, result);
    
  }


  // Create the Pipes for Li1->DSP1 (30 RawDataReceiver pipes)
  for (i = 0; i < g_demoEnv.coRawDevs; i++) {
    OTX_INT32 nTs = i%32;
    OTX_INT32 nDsp = g_hLogicDevs.mRawDataReceiver[i].nDsp;
    OTX_INT32 nLi = g_hLogicDevs.mRawDataReceiver[i].nLi;
        
    if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nTs] = 0xFF; 
    
    result = OtxDrvCreatePipe(nCapacity, &g_hLogicDevs.mRawDataReceiver[i].hPipe); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hPipe, result);
    
    // Connect pipe: Li  -> Dsp 
    result = OtxDrvConnectPipe(g_hLogicDevs.mRawDataReceiver[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hPipe, result);
    result = OtxDrvConnectPipe(g_hLogicDevs.mRawDataReceiver[i].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hPipe, result);
    
  }



  // CONNECT LOGICAL DEVICES WITHIN THE DSPs
  // ---------------------------------------
  for (i = 0; i < g_demoEnv.coRawDevs; i++) {
    if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)

    // Receive direction:
    // Connect Raw Data Receiver
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.mRawDataReceiver[i].hPipe,  -1,  g_hLogicDevs.mRawDataReceiver[i].hHandle, -1); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hPipe, result);
    // Transmit direction:
    // Connect Raw Data Sender
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.mRawDataSender[i].hHandle, -1, g_hLogicDevs.mRawDataSender[i].hPipe, -1); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);

  }

 
}


void DemoInitializeLogicalDevices(
/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after 
    the devices are created and connected.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < 2; i++) {

    // Initialize the Hdlc Sender and Receiver in the Lis
    result = OtxHdlcSenderInit(
      g_hLogicDevs.hHdlcSender[i], // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
      2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
      OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
      ); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);          
    
    result = OtxHdlcReceiverInit(
      g_hLogicDevs.hHdlcReceiver[i], // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
      2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
      OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
      OTX_TRUE                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
      ); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);          
    
  }

  for (i = 0; i < g_demoEnv.coRawDevs; i++) {
    if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
    // Initialize the Raw Data Receivers (hosted in the DSP)
    result = OtxRawDataReceiverInit(
      g_hLogicDevs.mRawDataReceiver[i].hHandle, 
      OTX_DATA_FORMAT_UINT8,  // No data conversion. OtxHdlc SPM only supports this format
      0xFF,        // Flag to be filtered
      OTX_FALSE,   // Disable Flag filtering
      OTX_TRUE     // Least Significant part is received first. Only applicable for a subchannel pipe (nCapacity < 64).
      ); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
    // Initialize the Raw Data Senders (hosted in the DSP)
    result = OtxRawDataSenderInit(
      g_hLogicDevs.mRawDataSender[i].hHandle, 
      OTX_DATA_FORMAT_UINT8,  // No data conversion. OtxHdlc SPM only supports this format
      0xFF,        // Flag to be sent when the sender is out of data
      OTX_TRUE     // Least Significant part is transmitted first. Only applicable for a subchannel pipe (nCapacity < 64).
      ); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
  }
}


void DemoEnableLogicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < 2; i++) {

    DemoLogf("Enable hHdlcSender[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.hHdlcSender[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);

    DemoLogf("Enable hHdlcReceiver[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.hHdlcReceiver[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);
  }


  for (i = 0; i < g_demoEnv.coRawDevs; i++) {
    if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)

    DemoLogf("Enable mRawDataSender[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);

    //DemoLogf("Enable mRawDataReceiver[%d]\n", i);
    //result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
  }


}


void DemoInitT1E1ConfigData(
/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 128;
  pE1CfgData->nLosRecovery               = 128;
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 128;
  pT1CfgData->nLosRecovery               = 128;
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x00;
}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (result == OTX_S_OK);
  }

  return 0;

}

/*******************************************************************************
  @func
    Raw Data Receiver Thread Function 
*******************************************************************************/
#ifdef DEMO_LINK_WITH_LIBCMT
void DemoRawDataReceiver(
  void *lpParameter   // thread data
)
#else
void *DemoRawDataReceiver(
  void *lpParameter   // thread data
)
#endif
{
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  OTX_UINT32 nFrameLen;
  OTX_BOOL bFrameEnd;
  char msgStr[256];
#if !defined(_WIN32_WCE)
  clock_t startTime, stopTime;
#endif  // #if !defined(_WIN32_WCE)

  OTX_UINT32 nSourceId;

#if !defined(_WIN32_WCE)
  startTime = clock();
#endif  // #if !defined(_WIN32_WCE)

  nSourceId = *((OTX_INT32 *) lpParameter);
  if (nSourceId >= (OTX_UINT32)g_demoEnv.coRawDevs) {
    DemoLogf("DemoRawDataReceiver: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    goto END;
  }

  g_hLogicDevs.mRawDataReceiver[nSourceId].cbReceived = 0;

/*
  // Discard any events from event queue (if any)
  do {
    result = OtxDrvGetEventData(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, result);
  } while (result == OTX_S_OK);
  // Discard any events from event queue (if any)
  do {
    result = OtxRawDataReadData(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, g_hLogicDevs.mRawDataReceiver[nSourceId].sqRxBuf, 2048, &nFrameLen, &bFrameEnd);
  } while (result == OTX_S_OK);
  if (result != OTX_S_FALSE) {
    DemoLogf("RawDataReceiver[%d] OtxRawDataReadData() was unsuccessful.\n", nSourceId);
    DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, result);
  }
*/


  while (1) {
    OtxDrvWaitForSingleEvent(g_hLogicDevs.mRawDataReceiver[nSourceId].hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        switch (eventData.m_eDeviceType) {
        case OTX_LDEVICE_DATA_RAW_RECEIVER:
          // Event from Raw Data Receiver 
          if ((eventData.m_nSourceId != nSourceId)) {
            DemoLogf("RawDataSender[%d] Error: m_nSourceId is out of range.\n", eventData.m_nSourceId);
            break;
          } else {
             i = eventData.m_nSourceId;
          }

          switch (eventData.m_nCode) {
          case OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE:
            {
              otxRc = OtxRawDataReadData(eventData.m_hDevice, g_hLogicDevs.mRawDataReceiver[nSourceId].sqRxBuf, 2048, &nFrameLen, &bFrameEnd);
              while (otxRc == OTX_S_OK) {
                
                //DemoLogf("RawDataSender[%d] Received data \"%s\" (len=%d, end=%d) \n", eventData.m_nSourceId, (char *)g_hLogicDevs.mRawDataReceiver[nSourceId].sqRxBuf, nFrameLen, bFrameEnd);

                g_hLogicDevs.mRawDataReceiver[nSourceId].cbReceived += nFrameLen;

                if (bFrameEnd) {
                  DemoLogf("RawDataReceiver[%d] Error: Unexpected FrameEnd received \n", eventData.m_nSourceId);
                }
                
                otxRc = OtxRawDataReadData(eventData.m_hDevice, g_hLogicDevs.mRawDataReceiver[nSourceId].sqRxBuf, 2048, &nFrameLen, &bFrameEnd);
              }
              if (otxRc != OTX_S_FALSE) {
                DemoLogf("RawDataReceiver[%d] OtxRawDataReadData() was unsuccessful.\n", eventData.m_nSourceId);
                //DemoCheckRc(eventData.m_hDevice, otxRc);
              }

            }
            break;
          default:
            otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
            // Unexpected event from this device
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("RawDataReceiver[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
            } else {
              DemoLogf("RawDataReceiver[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
            }
            break;
          }
          break;
          default:
            // Unhandled event
            DemoLogf("DemoRawDataSender[%d]: Event from an unknown object 0x%X\n", nSourceId, eventData.m_hDevice);
            break;
        }
        
      }
    } while (result == OTX_S_OK);

    OsaThreadSleep(256); // Give the driver some time to fill up more data for us to fetch (256ms fills 2048 bytes in a 64kbit/s pipe)

  }

END:
  // We should never get here
#if !defined(_WIN32_WCE)  
  stopTime = clock();
  
  DemoLogf("DemoRawDataReceiver[%d]: Done (time=%2.1f seconds).\n", nSourceId, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
#endif  // #if !defined(_WIN32_WCE)  
#ifdef DEMO_LINK_WITH_LIBCMT
  _endthread();
#else
  return 0;
#endif

}

/*******************************************************************************
  @func
    Raw Data Sender Thread Function 
*******************************************************************************/

#ifdef DEMO_LINK_WITH_LIBCMT
void DemoRawDataSender(
  void *lpParameter   // thread data
)
#else
void *DemoRawDataSender(
  void *lpParameter   // thread data
)
#endif
{
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];
#if !defined(_WIN32_WCE)
  clock_t startTime, stopTime;
#endif  // #if !defined(_WIN32_WCE)
  OTX_UINT32 nSourceId;

#if !defined(_WIN32_WCE)  
  startTime = clock();
#endif  // #if !defined(_WIN32_WCE)

  nSourceId = *((OTX_INT32 *) lpParameter);
  if (nSourceId >= (OTX_UINT32)g_demoEnv.coRawDevs) {
    DemoLogf("DemoRawDataSender: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    goto END;
  }

  g_hLogicDevs.mRawDataSender[nSourceId].coUnderRunErrors = 0;

  // Fill a buffer of a constant value to be sent by this sender
  memset(g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, nSourceId, sizeof(OTX_UINT8) * DEMO_RAW_TX_BUF_SIZE);
  g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf = DEMO_RAW_TX_BUF_SIZE;
  g_hLogicDevs.mRawDataSender[nSourceId].cbSent = 0;


  // Randomize the start to get all senders staggered (to even out the processor load)
  //OsaThreadSleep((rand() % 200) + 50); 


  // Discard any events from event queue (if any)
  do {
    result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, result);

  } while (result == OTX_S_OK);


  // Prime the driver buffer with data buffers to be sent (if the device is enabled)
  do {
    otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, DEMO_TASK_REF, OTX_TIME_INFINITY);
    if (result == OTX_S_OK) {
      g_hLogicDevs.mRawDataSender[nSourceId].cbSent += g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf;
    }
  } while ((otxRc != OTX_E_BUSY) && (otxRc != OTX_E_NOT_WORKING));
    
  
  while (1) {
    OtxDrvWaitForSingleEvent(g_hLogicDevs.mRawDataSender[nSourceId].hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        switch (eventData.m_eDeviceType) {
        case OTX_LDEVICE_DATA_RAW_SENDER:
          // Event from Raw Data Sender 
          if ((eventData.m_nSourceId != nSourceId)) {
            DemoLogf("RawDataSender[%d] Error: m_nSourceId is out of range.\n", eventData.m_nSourceId);
            break;
          } else {
            i = eventData.m_nSourceId;
            //DemoLogf(" %d", eventData.m_nSourceId);
          }
          
          switch (eventData.m_nCode) {
          case OTX_RAWDATA_SENDER_EC_COMPLETED:
            g_hLogicDevs.mRawDataSender[i].coUnderRunErrors++;
            //DemoLogf("DemoRawDataSender[%d]: *** Data Underrun at cbTxSent/cbTxBuf=%d/%d (m_nParam=%d). \n", nCin, g_sqCin[nCin].cbTxSent, g_sqCin[nCin].cbTxBuf, eventData.m_nParam);
            // Fall-through
          case OTX_RAWDATA_SENDER_EC_BUF_DEPLETED:
            {
              OTX_RESULT otxRc;
              OTX_INT32 nCount = 0;
              
              do {
                nCount++;
                otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[i].hHandle, g_hLogicDevs.mRawDataSender[i].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, DEMO_TASK_REF, OTX_TIME_INFINITY);
                if (otxRc == OTX_S_OK) {
                  g_hLogicDevs.mRawDataSender[nSourceId].cbSent += g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf;
                }
                //if (nCount > 1) DemoLogf(".");
              } while ((otxRc != OTX_E_BUSY) && (otxRc != OTX_E_NOT_WORKING));
              
              OsaThreadSleep(0); // Give the driver some time to empty its queue of data (128ms fills 1024 bytes in a 64kbit/s pipe)
              
            }
            break;
            
            
          default:
            otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
            // Unexpected event from this device
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("RawDataSender[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
            } else {
              DemoLogf("RawDataSender[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
              
            }
            break;
          }
          break;
        default:
          // Unhandled event
          DemoLogf("DemoRawDataSender[%d]: Event from an unknown object 0x%X\n", nSourceId, eventData.m_hDevice);
          break;
        }
        
      }
    } while (result == OTX_S_OK);
  }

END:
  // We should never get here
#if !defined(_WIN32_WCE)  
  stopTime = clock();
  
  DemoLogf("DemoRawDataSender[%d]: Done (time=%2.1f seconds, coUnderRunErrors=%d).\n", nSourceId, (double)(stopTime - startTime) / CLOCKS_PER_SEC, g_hLogicDevs.mRawDataSender[nSourceId].coUnderRunErrors);
#endif  // #if !defined(_WIN32_WCE)  
#ifdef DEMO_LINK_WITH_LIBCMT
  _endthread();
#else
  return 0;
#endif

}


/*******************************************************************************
  @func
    Print buffer usage statistics.
*******************************************************************************/
#ifdef DEMO_LINK_WITH_LIBCMT
void DemoStatsThread(
  void *lpParameter   // thread data
)
#else
void *DemoStatsThread(
  void *lpParameter   // thread data
)
#endif
{
  OTX_RESULT result;
  OTX_UINT32 nCurrentUsage;
  OTX_UINT32 nMaxUsage;
  OTX_UINT32 nLostEvents;
  OTX_UINT32 nQueuedCurrent;
  OTX_UINT32 nQueuedPeak;
  OTX_INT32 nDsp;

  while (g_bStatsThreadActive) {
    
    for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
      
      //result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      //DemoLogf("  Size (in words)         : %d\n", nAttribValue);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_CURRENT_USAGE, &nCurrentUsage); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      //DemoLogf("  Current usage (in words): %d\n", nAttribValue);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_MAX_USAGE, &nMaxUsage); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      //DemoLogf("  Max usage (in words)    : %d\n", nAttribValue);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_LOST_EVENTS, &nLostEvents); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_EXT_QUEUE_CURRENT_USAGE, &nQueuedCurrent); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_EXT_QUEUE_PEAK_USAGE, &nQueuedPeak); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      
      DemoLogf("Dsp[%d] curr=%.8d, max=%.8d, lost=%d, qcurr=%d, qpeak=%d\n", nDsp, nCurrentUsage, nMaxUsage, nLostEvents, nQueuedCurrent, nQueuedPeak);
      
    }

    OsaThreadSleep(2000); 
  }

  DemoLogf("DemoStatsThread() terminating.\n");

  return 0;

    
}



void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  OTX_UINT8  buf[2048];
  OTX_UINT32 nFrameLen;
  OTX_UINT32 nFrameSeq;
  OtxHdlcFrameEndCodeE nFrameEndCode;
  OTX_DATETIME nTimeStamp;
  char msgStr[256];

  //EnterCriticalSection(&g_analyzeCritSec);


  DemoLogf("%.8X%.8X ", pEventData->m_nTimeStamp.sUint32.nHigh, pEventData->m_nTimeStamp.sUint32.nLow);
    
  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  case OTX_LDEVICE_HDLC_RECEIVER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from HLDC Receiver 
      if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
        
        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
        while (otxRc == OTX_S_OK) {

          DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
          
          DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
          for (i = 0; i < nFrameLen && i < 1024; i++) {
            DemoLogf("%.2X ", buf[i]);
          }
          DemoLogf("\n");

          otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
        }
        if (otxRc != OTX_S_FALSE) {
          DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
          DemoCheckRc(pEventData->m_hDevice, otxRc);
        }

      } else {
        // Unexpected event from this device
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


  //LeaveCriticalSection(&g_analyzeCritSec);

}



/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_PCMCIA;
  envP->coDsp             = 2;
  envP->coRawDevs         = 10;
  
  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
            envP->coDsp      = 4;
#if 0
            envP->coRawDevs  = 2;
#else
            envP->coRawDevs  = 32;
#endif
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
            envP->coDsp      = 2;
            envP->coRawDevs  = 16;        // with 32, DSPs overrun
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {        // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') { // T1
            DemoLogf("Error: This application is not implemented for T1 mode %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
            //envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|M]      : Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board with Vidar-5x4-ASM\n");
  DemoLogf("                  M = Use Thor-2-PCMCIA Board\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  //EnterCriticalSection(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  //LeaveCriticalSection(&g_demoLogfCritSec);
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d): %s \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      } else {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d) Cannot translate this result code \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber);
      }
    }


    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  static OTX_INT32 nCurrentLi = 0;
  static OTX_INT32 iRawRangeStart = 1;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-1 Select active Li or Dsp, this Li/Dsp is then used for subsequent commands\n");
    DemoLogf(" e   Enable range of Raw Data Receivers\n");
    DemoLogf(" d   Disable range of Raw Data Receivers\n");
    DemoLogf(" E   Enable range of Raw Data Senders\n");
    DemoLogf(" D   Disable range of Raw Data Senders\n");
    DemoLogf(" r   Set range for Enable/Disable command [1..15, 17..31] [33..47, 49..63]\n");
    DemoLogf(" t   Send a HDLC frame on TS16 (using HDLC Sender in the Li)\n");
    DemoLogf(" l   Query Line Status of both Lis\n");
    DemoLogf(" v   Print Raw Data Receiver/Sender byte counter statistics\n");
    DemoLogf(" s   Toggle verbose buffer usage statistics\n");
    break;

#if 1 // Only for debugging
  case 'b':
    {
      char szBuf[12];
      OTX_INT32 nDsp;
      
      DemoLogf("Init Bsp. Which Dsp#?");
      OsaConsoleGetStr(szBuf);
      nDsp = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);

    }
    break;
  case 'B':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;

  case 'O':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Dump debug data for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 2);
      }
    }

#endif

  case 'E':
    {
      OTX_INT32 i;
      for (i = iRawRangeStart; i <= g_iRawRangeEnd; i++) {
        if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
        
        DemoLogf("Enable mRawDataSender[%d]\n", i);
        result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
      }
    }
    break;


  case 'e':
    {
      OTX_INT32 i;
      for (i = iRawRangeStart; i <= g_iRawRangeEnd; i++) {
        if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
        
        DemoLogf("Enable mRawDataReceiver[%d]\n", i);
        result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
      }
    }
    break;


  case 'i':
    {
      OTX_INT32 i;
      for (i = 0; i < g_demoEnv.coRawDevs; i++) {
        if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
        
        DemoLogf("Enable mRawDataReceiver[%d]\n", i);
        result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
        //OsaThreadSleep(2000);
      }
    }
    break;


  case 'I':
    {
      OTX_INT32 i;
      for (i = 0; i < g_demoEnv.coRawDevs; i++) {
        if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
        
        DemoLogf("Disable mRawDataReceiver[%d]\n", i);
        result = OtxDrvDisable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
        //OsaThreadSleep(2000);
      }
    }
    break;


  case 'j':
    {
      OTX_INT32 i;

      for (;;) {
        for (i = 0; i < g_demoEnv.coRawDevs; i++) {
          if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
          
          DemoLogf("Enable mRawDataReceiver[%d]\n", i);
          result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
          //OsaThreadSleep(2000);
        }
        
        for (i = 0; i < g_demoEnv.coRawDevs; i++) {
          if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
          
          DemoLogf("Disable mRawDataReceiver[%d]\n", i);
          result = OtxDrvDisable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
          //OsaThreadSleep(2000);
        }
      }


    }
    break;



  case 'D':
    {
      OTX_INT32 i;
      for (i = iRawRangeStart; i <= g_iRawRangeEnd; i++) {
        if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
        
        DemoLogf("Disable mRawDataSender[%d]\n", i);
        result = OtxDrvDisable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
      }
    }
    break;


  case 'd':
    {
      OTX_INT32 i;
      for (i = iRawRangeStart; i <= g_iRawRangeEnd; i++) {
        if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
        
        DemoLogf("Disable mRawDataReceiver[%d]\n", i);
        result = OtxDrvDisable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
      }
    }
    break;



  case 'f':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'F':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Reseting Li #%d... ", nLi);
        result = OtxDrvReset(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;


  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;



  case 'v':
    {
      OTX_INT32 i;
      OtxDeviceStateE  eDevState;

      for (i = 0; i < g_demoEnv.coRawDevs; i++) {
        DemoLogf("%.2d: ", i);
        DemoLogf("Tx[%d,%d]", g_hLogicDevs.mRawDataSender[i].nDsp, g_hLogicDevs.mRawDataSender[i].nLi);
        OtxDrvGetState(g_hLogicDevs.mRawDataSender[i].hHandle, &eDevState);
        DemoLogf("st=%d ", (int)eDevState);
        DemoLogf("cb=%d ", g_hLogicDevs.mRawDataSender[i].cbSent);
        DemoLogf("ure=%d ", g_hLogicDevs.mRawDataSender[i].coUnderRunErrors);
        DemoLogf(" Rx[%d,%d]", g_hLogicDevs.mRawDataReceiver[i].nDsp, g_hLogicDevs.mRawDataReceiver[i].nLi);
        OtxDrvGetState(g_hLogicDevs.mRawDataReceiver[i].hHandle, &eDevState);
        DemoLogf("st=%d ", (int)eDevState);
        DemoLogf("cb=%d ", g_hLogicDevs.mRawDataReceiver[i].cbReceived);
        DemoLogf("%.2X%.2X%.2X ", g_hLogicDevs.mRawDataReceiver[i].sqRxBuf[0], g_hLogicDevs.mRawDataReceiver[i].sqRxBuf[1], g_hLogicDevs.mRawDataReceiver[i].sqRxBuf[2]);
        DemoLogf("\n");
      }
      DemoLogf("\n");
    }
    break;

  case 't':
    {
      OTX_UINT8 arMsg[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
      OTX_INT32 i;
      OTX_INT32 busyCount = 0;
#if !defined(_WIN32_WCE)
      clock_t startTime, stopTime;
#endif  // #if !defined(_WIN32_WCE)

      if (g_cbLiHdlcSenderFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      for (i = 0; i < g_cbLiHdlcSenderFrame; i++) {
        arMsg[i] = (OTX_UINT8) i;
      }

      DemoLogf("Sending %d HDLC frame(s) (TS16) from Li #%d (len=%d) ...", g_coHdlcSenderFrames, nCurrentLi, g_cbLiHdlcSenderFrame);

      // Send HDLC Frame(s)
#if !defined(_WIN32_WCE)
      startTime = clock();
#endif  // #if !defined(_WIN32_WCE)
      for (i = 0; i < g_coHdlcSenderFrames; i++) {
        result = OtxHdlcSendData(g_hLogicDevs.hHdlcSender[nCurrentLi], arMsg, g_cbLiHdlcSenderFrame, DEMO_TASK_REF, OTX_TIME_INFINITY); 
        if (result == OTX_E_BUSY) {
          // Driver IO Request buffer is full. We need to retransmit this frame.
          i--;
          busyCount++;
        } else {
          DemoCheckRc(g_hLogicDevs.hHdlcSender[nCurrentLi], result);
        }
      }
#if !defined(_WIN32_WCE)
      stopTime = clock();

      DemoLogf("Done (BusyCount=%d, time=%2.1f seconds).\n", busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
#endif  // #if !defined(_WIN32_WCE)
    }
    break;

  case 'r':
    {
      char strBuf[12];
      DemoLogf("Range for Raw Data command\n");

      DemoLogf("Start index (%d) ", iRawRangeStart);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        iRawRangeStart = (OTX_INT32)strtol(strBuf, NULL, 0);
        if (iRawRangeStart%16 == 0) {
          iRawRangeStart++;
        }
        if (iRawRangeStart >= g_demoEnv.coRawDevs) {
          iRawRangeStart = g_demoEnv.coRawDevs - 1;
        }
      }

      DemoLogf("End index (%d) ", g_iRawRangeEnd);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_iRawRangeEnd = (OTX_INT32)strtol(strBuf, NULL, 0);
        if (g_iRawRangeEnd%16 == 0) {
          g_iRawRangeEnd--;
        }
        if (g_iRawRangeEnd >= g_demoEnv.coRawDevs) {
          g_iRawRangeEnd = g_demoEnv.coRawDevs - 1;
        }
        if (g_iRawRangeEnd < 0) {
          g_iRawRangeEnd = 1;
        }
      }

      DemoLogf("Range is now set to [%d, %d]\n", iRawRangeStart, g_iRawRangeEnd);
    }
    break;

  case 's':
    {
      g_bStatsThreadActive = !g_bStatsThreadActive;

      if (g_bStatsThreadActive) {
        DemoLogf("Starting Stats Thread\n");
        OsaThreadStart(DemoStatsThread);
      }

    }
    break;


  case '0':
    // Select Li/Dsp 0 as current
    nCurrentLi = 0;
    DemoLogf("Li/Dsp#0 Selected ....\n");
    break;

  case '1':
    // Select Li/Dsp 1 as current
    nCurrentLi = 1;
    DemoLogf("Li/Dsp#1 Selected ....\n");
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}                      






int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_INT32 i;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  DemoInitalizeGlobalData();

  DemoLogf("T2MRAW30, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Example program for Raw Data Sender/Receiver (30x64kpbs).\n");

  //InitializeCriticalSection(&g_analyzeCritSec);
  //InitializeCriticalSection(&g_demoLogfCritSec);


  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  
  {
    OTX_INT32  i;
    for (i = 0; i < DEMO_RAWDATA_DEVICES_MAX; i++) {
      DemoLogf("Rx[%.2d] nDsp=%d, nLi=%d,    Tx[%.2d] nDsp=%d, nLi=%d\n", 
        i, 
        g_hLogicDevs.mRawDataReceiver[i].nDsp,
        g_hLogicDevs.mRawDataReceiver[i].nLi,
        i, 
        g_hLogicDevs.mRawDataSender[i].nDsp,
        g_hLogicDevs.mRawDataSender[i].nLi
        );
    }
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board (important to set the clock source before configuring the E1/T1 devices)
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_NA, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("DemoCreateLogicalDevices\n");
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Initialize the logical devices. A logical device must be create and connected before it can be initialized.
  DemoLogf("DemoInitializeLogicalDevices\n");
  DemoInitializeLogicalDevices();

  // Enable logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnableLogicalDevices\n");
  DemoEnableLogicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  for (i = 0; i < g_demoEnv.coRawDevs; i++) {
    
    if (i%16 == 0) continue; // Skip 0 (sync), 16 (signalling)
    
        
    DemoLogf("Creating Raw Data Receiver[%d] Thread\n", i);
    g_hLogicDevs.mRawDataReceiver[i].nSourceId = i;
#ifdef DEMO_LINK_WITH_LIBCMT
    _beginthread(DemoRawDataReceiver, 0, (void *) &g_hLogicDevs.mRawDataReceiver[i].nSourceId);
#else
  	OsaThreadStartEx(DemoRawDataReceiver, (void *) &g_hLogicDevs.mRawDataReceiver[i].nSourceId);
#endif


    DemoLogf("Creating Raw Data Sender[%d] Thread\n", i);
    g_hLogicDevs.mRawDataSender[i].nSourceId = i;
#ifdef DEMO_LINK_WITH_LIBCMT
    _beginthread(DemoRawDataSender, 0, (void *) &g_hLogicDevs.mRawDataSender[i].nSourceId);
#else
  	OsaThreadStartEx(DemoRawDataSender, (void *) &g_hLogicDevs.mRawDataSender[i].nSourceId);
#endif

    //OsaThreadSleep(200);
  }    
  
  g_iRawRangeEnd = g_demoEnv.coRawDevs-1;

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }


  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  static char exename[_MAX_PATH * sizeof(TCHAR)] = "";
  static char *cmdParams[] = { exename, "/ME", NULL }; 
  char *pc;

  GetModuleFileName(NULL, (LPTSTR) cmdParams[0], _MAX_PATH);
  TCHARTOSTR((LPTSTR) cmdParams[0], cmdParams[0]);
  strcpy(g_demoEnv.cwd, cmdParams[0]);
  pc = strrchr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
