/* $Workfile: OtxDsp.h $   $Revision: 22 $    $Modtime: 8/05/03 5:13p $  */
#ifndef OTX_DSP_H
#define OTX_DSP_H
/*******************************************************************************

  @doc PublicHwApi OtxDsp

  Copyright (C) 1998, 1999 Odin TeleSystems Inc., All Rights Reserved

  OTX Generic Hardware Driver API

  @module
  OtxDsp | This file contains prototypes for the DSP Management Functions

  @contents2
  OtxDsp.h |

  @index func | OtxDsp

  @contents2
  OTX DSP Device Management Functions - OtxDspXXX() | The user functions available for
  the Management of DSP Devices

  @head3 OtxDspXXX() Function Reference |

  @index func | OtxDsp

*******************************************************************************/

#include "OtxType.h"    // OTX specific type definitions
#include "OtxDef.h"     // OTX specific definitions

#include "OtxDspD.h"    // OTX DSP type definitions

#ifdef __cplusplus
extern "C" {
#endif


/*******************************************************************************

  @func
    Load and Run a program on a DSP (from a file)
  @xref
    <f OtxDspRunSpm>
    <f OtxDspRestartProgram>

*******************************************************************************/
OTX_RESULT OtxDspRunProgram(
  IN OTX_HANDLE hDspDevice,         // @parm Handle to DSP device
  IN const char *szPrgfile,         // @parm Application file to be loaded
  IN const char *szEntryPoint       // @parm Label for the entry point of the application
);


/*******************************************************************************

  @func
    Load and Run an OTX Standard Program Module (SPM) on a DSP from the a 
    program image stored in the lib.
  @xref
    <f OtxDspRunProgram>

*******************************************************************************/
OTX_RESULT OtxDspRunSpm(
  IN OTX_HANDLE hDspDevice,         // @parm Handle to DSP device
  IN OtxDspSpmE eProgramModule      // @parm Standard Program Module to be loaded in the DSP (from the Lib)
);



/*******************************************************************************

  @func
    Converts a DSP Application file in COFF format (*.out) for a C548/549 DSP to
    the OTX DSP Image format (used by <f OtxDspRunImage>). 
  @xref
    <f OtxDspRunImage>
    <f OtxDspRunProgram>

*******************************************************************************/
OTX_RESULT OtxDspCreateProgramImage(
  IN const char *szProgfile,            // @parm Application file to be converted to an image   
  IN const char *szEntryPoint,          // @parm Label for the entry point of the application. Usually "_c_int00" for a C-compiled application.
  IN OTX_UINT16 coProgImageMax,         // @parm Count of objects (size) of psqProgImage (in elements) provided by the user
  IN OUT OTX_UINT16 *psqProgImage,      // @parm Image Buffer provided by the user
  OUT OTX_UINT16 *coProgImage,          // @parm Number of image elements returned in psqProgImage 
  OUT OTX_INT32 *nEntryAddr             // @parm Program Entry address
);


/*******************************************************************************

  @func
    Converts a DSP Application file in COFF format (*.out) for 
    C5400 and c5510 DSPs to
    the OTX DSP Extended Image format (used by <f OtxDspRunImage>).
  @xref
    <f OtxDspRunImage>
    <f OtxDspRunProgram>

*******************************************************************************/
OTX_RESULT OtxDspCreateProgramImageEx(
  IN const char *szProgfile,            // @parm Application file to be converted to an image   
  IN const char *szEntryPoint,          // @parm Label for the entry point of the application. Usually "_c_int00" for a C-compiled application.
  IN OTX_UINT32 coProgImageMax,         // @parm Count of objects (size) of psqProgImage (in elements) provided by the user
  IN OUT OTX_UINT16 *psqProgImage,      // @parm Image Buffer provided by the user
  OUT OTX_UINT32 *coProgImage,          // @parm Number of image elements returned in psqProgImage 
  OUT OTX_INT32 *nEntryAddr             // @parm Program Entry address
);




/*******************************************************************************

  @func
    Loads and starts a DSP application program. The DSP application
    program must be in the OTX DSP Image format (created by a
    call to <f OtxDspCreateProgramImage>). 
  @xref
    <f OtxDspCreateProgramImage>

*******************************************************************************/
OTX_RESULT OtxDspRunImage(
  IN OTX_HANDLE  hDspDevice,      // @parm Handle to DSP device
  IN OTX_UINT16 *psqProgImage,    // @parm Program image 
  IN OTX_UINT16  coProgImage,     // @parm Number of elements in psqProgImage 
  IN OTX_INT32   nEntryAddr       // @parm Program Entry Address 
);

/*******************************************************************************

  @func
    DSP Debugging utility function. Restarts a previously loaded and/or 
    running program. 
    
    This function is only meant to be called if a DSP Debugger is being
    used. 

  @ex This function should only be used on the following call sequence: |

    1. First the OtxDspRunImage(), OtxDspRunSpm(), or OtxDspRunProgram() is called.
    2. Immediately thereafter the host application must be stopped.
    3. While the host application is stopped the DSP program is loaded and started 
       from the DSP Debugger.
    4. Once the DSP program is running again, OtxDspRunProgram() should be called
       to bring the DSP program back to the state it was right after step 1 was
       executed.

  @xref
    <f OtxDspRunProgram>

*******************************************************************************/
OTX_RESULT OtxDspRestartProgram(
  IN OTX_HANDLE hDspDevice          // @parm Handle to DSP device
);


/*******************************************************************************

  @func
    Send a user command to a user supplied DSP application
  @xref

*******************************************************************************/
OTX_RESULT OtxDspIoControl(
  IN  OTX_HANDLE hDspDevice,         // @parm Handle to the physical DSP device
  IN  OTX_UINT16 nIoControlCode,     // @parm I/O Control Code to be passed 
  IN  void       *pInBuf,            // @parm Pointer to the user allocated input buffer
  IN  OTX_UINT32 nInBufSize,         // @parm Size of the input buffer
  IN  void       *pOutBuf,           // @parm Pointer to the user allocated output buffer
  IN  OTX_UINT32 nOutBufSize,        // @parm Size of the output buffer
  OUT OTX_UINT32 *pnBytesReturned,   // @parm Number of Bytes returned
  IN  OTX_TASK_REF nTaskRef,         // @parm See <t OTX_TASK_REF>
  IN  OTX_TIME nWaitMaxMs            // @parm See <t OTX_TIME>
);



/*******************************************************************************

  @func
    Read data from a user supplied DSP application

  @rdesc Returns one of the following values:

    @flag OTX_S_OK | Data was fetched
    @flag OTX_S_FALSE | No Data available at this moment
    @flag OTX_E_BUFFER_TO_SMALL | The supplied user data buffer is too small to fit the received data

  @xref

*******************************************************************************/
OTX_RESULT OtxDspReadData(
  IN  OTX_HANDLE hDspDevice,         // @parm Handle to the physical DSP device
  IN  OTX_TASK_REF nTaskRef,         // @parm See <t OTX_TASK_REF>
  OUT OTX_UINT8  *pOutBuf,           // @parm Pointer to the user allocated output buffer
  IN  OTX_UINT32 nOutBufSize,        // @parm Size of the output buffer
  OUT OTX_UINT32 *pnBytesReturned,   // @parm Number of Bytes returned
  OUT OTX_UINT32 *pnIoControlCode    // @parm I/O Control Code the returned data is associated with
);


/*******************************************************************************

  @func
    Read a number of octets from the DSP's Host Port Interface (HPI)
    into a user supplied buffer.

   @comm
    Note: This function should be used with caution. It is intended for
    debugging purposes and advanced users only.

  @xref
    <f OtxDspWriteHpi>

*******************************************************************************/
OTX_RESULT OtxDspReadHpi(
  IN  OTX_HANDLE hDspDevice,         // @parm Handle to the physical DSP device
  IN  OTX_UINT32 nHpiAddress,        // @parm HPI Address to start to read from 
  IN  OTX_UINT32 coWordsToRead,      // @parm Number of OTX_UINT16s (Words) to read from the HPI
  IN  OUT OTX_UINT16 *pReadBuf       // @parm Pointer to the user allocated output buffer (Words)
);


/*******************************************************************************

  @func
    Write a number of octets to the DSP's Host Port Interface (HPI)
    from a user supplied buffer.

   @comm
    Note: This function should be used with caution since it can corrupt the
    communication buffers between the DSP and the host. It is intended for
    debugging purposes and advanced users only.

  @xref
    <f OtxDspReadHpi>

*******************************************************************************/
OTX_RESULT OtxDspWriteHpi(
  IN  OTX_HANDLE hDspDevice,         // @parm Handle to the physical DSP device
  IN  OTX_UINT32 nHpiAddress,        // @parm HPI Address to start to write to
  IN  OTX_UINT32 coWordsToWrite,     // @parm Number of OTX_UINT16s (Words) to write to the HPI
  IN  OUT OTX_UINT16 *pWriteBuf      // @parm Pointer to the user allocated input buffer (Words)
);



#ifdef __cplusplus
}  // extern "C"
#endif

/*******************************************************************************

  Copyright (C) Odin TeleSystems Inc., 1998, 1999

  All rights reserved. No part of this computer program may be reproduced,
  stored in any retrieval system, or transmitted, in any form or by any
  means, electronic, mechanical, photocopying, recording, or otherwise,
  without the prior written permission of Odin TeleSystems Inc., or without
  a permission stipulated in the agreement/contract under which this computer
  program was supplied.

*******************************************************************************/

#endif    // OTX_DSP_H
